/**
 * This class instantiates an RPN Calculator that reads input from
 * standard in and then computes and displays the answer or prints
 * an error message.
 *
 * @author <A HREF="mailto:gte220j@prism.gatech.edu.com"> J.Chris Holmes </A>
 * @version Version 1.0, 4/23/2001
 */

import java.util.Stack;
import java.util.StringTokenizer;
import java.io.IOException;

public class RPN_Calc
{

  /**
   * Default constructor for an RPN Calculator object
   */
  public RPN_Calc() {
    boolean not_done = true;
    String input_string; 
    int output;

    while( not_done == true) {
      try {
        input_string = IOHelp.readLine();

        if(input_string.compareToIgnoreCase("q") == 0) 
          not_done = false;
        else { 
          try {
            output = RunCalculation(input_string);
            System.out.println("ans = "+output);
          }
          catch(InvalidInputException iie) {
            System.out.println(iie.getMessage());
          }
        }
      }
      catch(IOException ioe) {
        ioe.printStackTrace();
        not_done = false;
      }
    }
  }

  /**
   * Takes in a String of math in Reverse Polish Notation form and
   * returns the answer or throws an Exception.
   *
   * @param input_string  A string of math in Reverse Polish Notation form
   * @return  An int of the result
   */
  public int RunCalculation(String input_string) throws InvalidInputException {
    StringTokenizer strtok = new StringTokenizer(input_string, " \t");
    Stack myStack = new Stack();
    String token;
    int op1, op2;

    while( strtok.hasMoreTokens() == true) {
      token = strtok.nextToken();

      if( isNum(token) == true) {
        myStack.push(token);
      }
      else if( isOperator(token) == true) {
        switch(token.charAt(0)) {
          case '+': if(myStack.size() < 2) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    op2 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+(op1+op2));
           break;
          case '-': if(myStack.size() < 2) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    op2 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+(op1-op2));
           break;
          case '*': if(myStack.size() < 2) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    op2 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+(op1*op2));
           break;
          case '/': if(myStack.size() < 2) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    op2 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+(op1/op2));
           break;
          case '~': if(myStack.size() < 1) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+(0-op1));
           break;
          case '^': if(myStack.size() < 2) 
	              throw new InvalidInputException("ERROR:  Expression Invalid");
                    op1 = Integer.parseInt((String)myStack.pop());
                    op2 = Integer.parseInt((String)myStack.pop());
                    myStack.push(""+((int)Math.pow((double)op1,(double)op2)));
           break;
          default: break;
        }
      }
      else {
        throw new InvalidInputException("ERROR:  Expression Invalid");
      }
    }   
    if(myStack.size() < 1) {
      throw new InvalidInputException("ERROR:  Expression Invalid");
    }
    return Integer.parseInt((String)myStack.pop());
  } 

  /**
   * Takes in a single string token and returns a boolean based on
   * whether or not the string is a valid number
   *
   * @param to_check  Any String
   * @return  A boolean that is true is the input String is a valid number
   */
  private boolean isNum(String to_check) {
    boolean toret = false;
    try {
      Integer.parseInt(to_check);
      toret = true;
    }
    catch(NumberFormatException nfe) {
      toret = false;
    }
    return(toret);
  }
  /**
   * Takes in a single string token and returns a boolean based on
   * whether or not the string is a valid mathematical operator
   *
   * @param to_check  Any String
   * @return  A boolean that is true is the input String is a valid operator
   */
  private boolean isOperator(String to_check) {
    if(to_check == null) return false;
    if(to_check.length() != 1) return false; 

    boolean toret = false;
    switch(to_check.charAt(0)) {
      case '+': 
      case '-':
      case '*':
      case '/':
      case '~':
      case '^':
        toret = true; break;
      default: break;
    }
    return(toret);
  }

  /**
   * Main method -- instantiates a Calculator
   *
   * @param args  Command line arguements
   */
  public static void main(String args[]) {
    new RPN_Calc();
  }
}
